# -*- coding: utf-8 -*-
"""
This module contains functions and classes related to post processing processes.
The post processing module is an addition to the currently available post processing
module of OpenSeesPy - this module fills in gaps to
* create envelope from xarray DataSet
* plot force and deflection diagrams from xarray DataSets
"""

import matplotlib.pyplot as plt
import opsvis as opsv
import numpy as np
from typing import TYPE_CHECKING, Union
from scipy.interpolate import interpn, RegularGridInterpolator

# if TYPE_CHECKING:
from ospgrillage.load import ShapeFunction
from ospgrillage.utils import solve_zeta_eta

import vfo.vfo as opsplt


def create_envelope(**kwargs):
    """
    Interface for users to create an :class:`Envelope` object.

    The constructor takes an `xarray` DataSet and kwargs for enveloping options.

    :param ds: Data set from `get_results()`
    :type ds: Xarray
    :param kwargs: Keyword arguments see below.

    :keyword:

    * array: either 'displacement' or 'forces'
    * value_mode: - True or False
    * query_mode: - True or False
    * extrema: either "min" or "max"
    * elements:
    * nodes
    * array
    * load_effect

    :return: :class:`Envelope` Object
    """
    return Envelope(**kwargs)


class Envelope:
    """
    Class for defining envelope of :class:`~ospgrillage.osp_grillage.OspGrillage`'s
    `xarray` of result.

    A :func:`Envelope.get` method is provided that returns an enveloped `xarray` based
    on the specified input parameters of this class.


    """

    def __init__(self, ds, load_effect: str = None, **kwargs):
        """

        :param ds: Data set from
                   :func:`~ospgrillage.osp_grillage.OspGrillage.get_results` . note Combination
        :type ds: Xarray
        :param load_effect: Specific load effect to envelope.
        :type load_effect: str
        :param kwargs: See below for keyword arguments.

        :keyword:

        * array: either 'displacement' or 'forces'
        * value_mode (`Bool`): Flag for envelope to return raw values - default True
        * query_mode (`Bool`): Flag for envelope to return loadcase coordinate for
                               maxima - default False
        * extrema (`str`): either "min" or "max"
        * elements:
        * nodes
        * array
        * load_effect

        """
        self.value = True
        self.ds = ds
        if ds is None:
            return

        # instantiate variables
        self.load_effect = (
            load_effect  # array load effect either displacements or forces
        )
        self.envelope_ds = None
        self.format_string = None
        # main command strings

        self.eval_string = (
            'self.ds.{array}.{xarray_command}(dim="Loadcase").sel({component_command})'
        )
        self.component_string = "Component={},"
        self.element_string = "Element={},"
        self.load_case_string = ""
        self.component_command = ""  # instantiate command string
        # default xarray function name
        self.xarray_command = {
            "query": ["idxmax", "idxmin"],
            "minmax value": ["max", "min"],
            "index": ["argmax", "argmin"],
        }
        self.selected_xarray_command = []
        # get keyword args
        self.elements = kwargs.get(
            "elements", None
        )  # specific elements to query/envelope
        self.nodes = kwargs.get("nodes", None)  # specific nodes to query/envelope
        self.component = kwargs.get(
            "load_effect", None
        )  # specific load effect to query
        self.array = kwargs.get("array", "displacements")
        self.value_mode = kwargs.get("value_mode", True)
        self.query_mode = kwargs.get("query_mode", False)  # default query mode
        self.extrema = kwargs.get("extrema", "max")

        # check variables
        if self.load_effect is None:
            raise Exception(
                "Missing argument for load_effect=: Hint requires a namestring of load"
                "effect type based on the Component dimension of the ospgrillage data"
                "set result format"
            )

        # process variables
        self.extrema_index = 0 if self.extrema == "max" else 1  # minima
        if self.query_mode:
            self.selected_xarray_command = self.xarray_command["query"][
                self.extrema_index
            ]
        elif self.value_mode:
            self.selected_xarray_command = self.xarray_command["minmax value"][
                self.extrema_index
            ]
        else:  # default to argmax/ argmin
            self.selected_xarray_command = self.xarray_command["index"][
                self.extrema_index
            ]

        # convert to lists
        if not isinstance(self.elements, list):
            self.elements = [self.elements]
        if not isinstance(self.component, list):
            self.component = [self.component]

        # check if empty
        if None not in self.elements:
            self.element_string.format(self.elements)
        if None not in self.component:
            self.component_string.format(self.component)

        # check the combinations of inputs for query
        if not ("{" in self.element_string and "{" in self.component_string):
            self.component_command = self.component_string + self.element_string
        elif "{" in self.element_string and "{" not in self.component_string:
            self.component_command = self.component_string
        elif not ("{" in self.element_string and "{" in self.component_string):
            self.component_command = self.element_string

        # format xarray command to be eval()
        self.format_string = self.eval_string.format(
            array=self.array,
            xarray_command=self.selected_xarray_command,
            component_command=self.component_command,
        )

    def get(self):
        """

        :return: The enveloped `xarray` object.
        :rtype: xarray
        """

        return eval(self.format_string)


def plot_force(
    ospgrillage_obj,
    result_obj=None,
    component=None,
    member: str = None,
    option: str = "elements",
    loadcase: str = None,
):
    """
    Plots a force diagram of the provided :class:`~ospgrillage.osp_grillage.OspGrillage` and
    :class:`xarray` (result) objects for a specified `component` and
    :class:`~ospgrillage.load.LoadCase`'s.

    .. note::
        For "shell_beam" model type, the function only plots the force diagrams for beam elements only.


    :param ospgrillage_obj: Grillage model object
    :type ospgrillage_obj: OspGrillage
    :param result_obj: xarray DataSet of results
    :type result_obj: xarray DataSet
    :param component: Force component to plot
    :type component: str
    :param member: member
    :type member: str
    :param option:
    :type option: str
    :param loadcase: name string of load case to plot. If not provided, plots from first load case in the order of
                     xarray loadcase coordinate
    :type loadcase: str
    :return: Matplotlib figure
    :rtype: (:class:`~matplotlib.figure.Figure`)
    """
    # instantiate component dict
    comp_dict = {"Fx": 0, "Fy": 1, "Fz": 2, "Mx": 3, "My": 4, "Mz": 5}
    comp_factor = {"Fx": 1, "Fy": 1, "Fz": 1, "Mx": 1, "My": 1, "Mz": -1}
    if member is None:
        print("Missing argument member=")
        return
    component_index = component
    if not isinstance(component, int):
        component_index = comp_dict[component]
    factor = comp_factor[component]
    fig, ax = plt.subplots()  # create plot window
    nodes = ospgrillage_obj.get_nodes()  # extract node information of model
    eletag = ospgrillage_obj.get_element(
        member=member, options=option
    )  # get ele tag of grillage elements
    # loop ele tags of ele
    if ospgrillage_obj.model_type == "shell_beam":
        force_result = result_obj.forces_beam
    else:
        force_result = result_obj.forces

    for ele in eletag:
        # get force components
        if loadcase:
            ele_components = force_result.sel(
                Loadcase=loadcase,
                Element=ele,
                Component=[
                    "Vx_i",
                    "Vy_i",
                    "Vz_i",
                    "Mx_i",
                    "My_i",
                    "Mz_i",
                    "Vx_j",
                    "Vy_j",
                    "Vz_j",
                    "Mx_j",
                    "My_j",
                    "Mz_j",
                ],
            ).values
        else:
            ele_components = force_result.sel(
                Element=ele,
                Component=[
                    "Vx_i",
                    "Vy_i",
                    "Vz_i",
                    "Mx_i",
                    "My_i",
                    "Mz_i",
                    "Vx_j",
                    "Vy_j",
                    "Vz_j",
                    "Mx_j",
                    "My_j",
                    "Mz_j",
                ],
            )[0].values
        # get nodes of ele
        if ospgrillage_obj.model_type == "shell_beam":
            ele_node = result_obj.ele_nodes_shell.sel(Element=ele)
            if all(np.isnan(result_obj.ele_nodes_shell.sel(Element=ele))):
                ele_node = result_obj.ele_nodes_beam.sel(Element=ele)
            # remove nans elements in the 4 node list (shell) for beam (2 elements)
            ele_node = ele_node[~np.isnan(ele_node)]
        else:
            ele_node = result_obj.ele_nodes.sel(Element=ele)

        # create arrays for x y and z for plots
        xx = [nodes[n]["coordinate"][0] for n in ele_node.values]
        yy = [nodes[n]["coordinate"][1] for n in ele_node.values]
        zz = [nodes[n]["coordinate"][2] for n in ele_node.values]
        # use ops_vis module to get force distribution on element
        # opsvis >= 1.0 replaced separate ex/ey/ez args with a single ecrd
        # ndarray of shape (2, 3): rows are nodes, columns are x, y, z
        s, *_ = opsv.section_force_distribution_3d(
            ecrd=np.column_stack([xx, yy, zz]),
            pl=ele_components,  # TODO check inputs with opsv package
        )

        # plot element force component
        ax.plot(xx, s[:, component_index] * factor, "-k")
        # fill area between horizontal axis and line
        ax.fill_between(
            xx, s[:, component_index] * factor, [0, 0], color="k", alpha=0.4
        )
    ax.set_title(member)
    ax.set_xlabel("x (m) ")
    ax.set_ylabel(component)
    fig.tight_layout()
    fig.show()

    return fig


def plot_defo(
    ospgrillage_obj,
    result_obj=None,
    member: str = None,
    component: str = None,
    option: str = "nodes",
    loadcase: str = None,
):
    """
    Plots displacements of the provided :class:`~ospgrillage.osp_grillage.OspGrillage` and
    :class:`xarray` (result) objects for a specified `component` and
    :class:`~ospgrillage.load.LoadCase`'s.

    .. note::
        For "shell_beam" model type, the function only plots the force diagrams for beam elements only.


    :param ospgrillage_obj: Grillage model object
    :type ospgrillage_obj: OspGrillage
    :param result_obj: xarray DataSet of results
    :type result_obj: xarray DataSet
    :param component: Force component to plot
    :type component: str
    :param member: member
    :type member: str
    :param option: option of :func:`~ospgrillage.osp_grillage.OspGrillage.get_element`,
                   either "nodes" or "element" (Default nodes)
    :type option: str
    :param loadcase: name string of load case to plot. If not provided, plots from first load case in the order of
                 xarray loadcase coordinate
    :type loadcase: str
    :return: Matplotlib figure
    :rtype: (:class:`~matplotlib.figure.Figure`)
    """
    # init vars
    previous_def = None
    previous_xx = None
    previous_zz = None

    # check options
    if option is None:
        plot_option = "nodes"
    else:
        plot_option = option
    # check member, if None, return None, Users need to define the member str to plot
    if member is None:
        print("Missing argument for member= - no plot is returned")
        return
    # check if component is provided, else default to
    dis_comp = component
    if component is None:
        dis_comp = "y"  # default to dy

    fig, ax = plt.subplots()
    # get all node information
    nodes = ospgrillage_obj.get_nodes()  # dictionary containing information of nodes
    # get specific nodes for specific element
    nodes_to_plot = ospgrillage_obj.get_element(member=member, options=plot_option)[
        0
    ]  # list of list
    # loop through nodes to plot
    for node in nodes_to_plot:
        if loadcase:
            disp = result_obj.displacements.sel(
                Component=dis_comp, Node=node, Loadcase=loadcase
            ).values  # get node disp value
        else:
            disp = result_obj.displacements.sel(Component=dis_comp, Node=node)[
                0
            ].values  # get node disp value
        xx = nodes[node]["coordinate"][0]  # get x coord
        zz = nodes[node]["coordinate"][2]  # get z coord (for 3D plots)
        if previous_def is not None:
            ax.plot(
                [previous_xx, xx], [previous_def, disp], "-b"
            )  # plot a 1-D plot of all nodes in grillage element
        previous_def = disp
        previous_xx = xx
        previous_zz = zz
    ax.set_title(member)
    ax.set_xlabel("x (m) ")  # labels
    ax.set_ylabel(dis_comp)  # labels
    fig.tight_layout()
    # fig.show()

    return fig


class PostProcessor:
    """Class to post-process the results from an of ospgrillage analysis result.

    As of version 0.2.0, ospgrillage compiles results using Xarray module.
    """

    def __init__(self, grillage, result):  # Union[xr.DataArray, xr.Dataset]
        """Init the class"""
        # store main vars
        self.grillage = grillage
        self.result = result

        # init vars
        self.shape_function_obj = ShapeFunction()

    def get_arbitrary_displacements(
        self, point: list, shape_function_type: str = "linear", component: str = "y"
    ):
        """Returns displacement (translation and rotational) from an arbitrary point by interpolation

        param point: list of coordinate. Default three elements [x,y=0,z]
        type point: list
        param component: Displacement component. Default "y"
        type component: str
        param shape_function_type: The shape function for interpolation. Default "linear"
        type shape_function_type: str
        """
        node_displacements = []
        node_coordinate = []

        # get the list of four nodes where arbitrary point lies
        nodes, grid_number = self.grillage._get_point_load_nodes(
            point=point
        )  # list of nodes
        if nodes is None:
            raise ValueError("Point is outside bridge mesh")

        # get results of each node of four nodes
        for node in nodes:
            node_displacements.append(
                self.result.sel(
                    Component=component,
                    Node=node,
                )
                .displacements.to_numpy()
                .tolist()[0]
            )
            node_coordinate.append(self.grillage.get_nodes(number=node))

        # interpolate for vertical displacement
        x = np.array([coord[0] for coord in node_coordinate])
        z = np.array([coord[2] for coord in node_coordinate])

        # get natural coordinate of point in grid
        eta, zeta = solve_zeta_eta(
            xp=point[0],
            zp=point[2],
            x1=x[0],
            z1=z[0],
            x2=x[1],
            z2=z[1],
            x3=x[2],
            z3=z[2],
            x4=x[3],
            z4=z[3],
        )
        if shape_function_type is "linear":
            shape_func = self.shape_function_obj.linear_shape_function(
                eta=eta, zeta=zeta
            )
        else:
            shape_func, _, _ = self.shape_function_obj.hermite_shape_function_2d(
                eta=eta, zeta=zeta
            )

        return sum([a * b for a, b, in zip(shape_func, node_displacements)])



import sys
import os
from PyQt5.QtWidgets import (QApplication, QMainWindow, QWidget, QTabWidget, 
                            QVBoxLayout, QHBoxLayout, QFormLayout, QGroupBox,
                            QLineEdit, QDoubleSpinBox, QSpinBox, QComboBox,
                            QPushButton, QLabel, QScrollArea, QMenuBar, 
                            QToolBar, QAction, QStatusBar, QTextEdit, QCheckBox,
                            QMessageBox, QRadioButton, QFileDialog)
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon

class BridgeInputWidget(QWidget):
    def __init__(self):
        super().__init__()
        self.setup_ui()
        
    def setup_ui(self):
        # Main layout
        main_layout = QVBoxLayout()
        main_layout.setContentsMargins(10, 10, 10, 10)
        
        # Create tab widget for different input categories
        self.tabs = QTabWidget()
        
        # Add input tabs
        self.create_geometry_tab()
        self.create_materials_tab()
        self.create_sections_tab()
        self.create_members_tab()
        #self.create_loads_tab()
        #self.create_analysis_tab()
        
        # Add tabs to main layout
        main_layout.addWidget(self.tabs)
        
        # Add control buttons
        button_layout = QHBoxLayout()
        self.btn_apply = QPushButton("Apply Changes")
        self.btn_run = QPushButton("Create Geometry")
        button_layout.addWidget(self.btn_apply)
        button_layout.addWidget(self.btn_run)
        main_layout.addLayout(button_layout)
        
        self.setLayout(main_layout)
    
    def create_geometry_tab(self):
        """Geometry input tab"""
        tab = QWidget()
        layout = QVBoxLayout()
        
        # Basic Geometry Group (unchanged)
        basic_group = QGroupBox("Basic Geometry")
        basic_form = QFormLayout()
        
        self.bridge_name = QLineEdit("My Bridge")
        self.bridge_length = QDoubleSpinBox()
        self.bridge_length.setRange(1, 1000)
        self.bridge_length.setValue(30)
        self.bridge_length.setSuffix(" m")
        
        self.bridge_width = QDoubleSpinBox()
        self.bridge_width.setRange(1, 100)
        self.bridge_width.setValue(10)
        self.bridge_width.setSuffix(" m")
        
        self.left_skew_angle = QDoubleSpinBox()
        self.left_skew_angle.setRange(-89, 89)
        self.left_skew_angle.setValue(0)
        self.left_skew_angle.setSuffix(" °")
        
        self.right_skew_angle = QDoubleSpinBox()
        self.right_skew_angle.setRange(-89, 89)
        self.right_skew_angle.setValue(0)
        self.right_skew_angle.setSuffix(" °")
        
        basic_form.addRow("Bridge Name", self.bridge_name)
        basic_form.addRow("Length", self.bridge_length)
        basic_form.addRow("Width", self.bridge_width)
        basic_form.addRow("Left Skew Angle", self.left_skew_angle)
        basic_form.addRow("Right Skew Angle", self.right_skew_angle)
        basic_group.setLayout(basic_form)
        
        # Mesh Settings Group
        mesh_group = QGroupBox("Mesh Settings")
        mesh_form = QFormLayout()
        
        self.bridge_type = QComboBox()
        self.bridge_type.addItems(["Straight", "Multi-Span Straight", "Curved"])
        
        
        self.long_beams = QSpinBox()
        self.long_beams.setRange(2, 20)
        self.long_beams.setValue(5)
        
        self.trans_beams = QSpinBox()
        self.trans_beams.setRange(2, 100)
        self.trans_beams.setValue(10)
        
        self.mesh_type = QComboBox()
        self.mesh_type.addItems(["Ortho", "Oblique"])

        self.opensees_file = QRadioButton("Opensees Command File")
        self.visualize = QRadioButton("Visualization")
        self.visualize.setChecked(True)
        
        # Add widgets to form
        mesh_form.addRow("Bridge Type", self.bridge_type)
##        mesh_form.addRow("Radius", self.bridge_radius)  # Added but hidden
        mesh_form.addRow("Longitudinal Beams", self.long_beams)
        mesh_form.addRow("Transverse Beams", self.trans_beams)
        mesh_form.addRow("Mesh Type", self.mesh_type)
        mesh_form.addRow(self.opensees_file)
        mesh_form.addRow(self.visualize)
        mesh_group.setLayout(mesh_form)

        # Create radius input but hide it initially
        self.radius_label = QLabel("Radius:")  # Store as instance variable
        self.bridge_radius = QDoubleSpinBox()
        self.bridge_radius.setRange(10, 10000)  # Adjust range as needed
        self.bridge_radius.setValue(100)
        self.bridge_radius.setSuffix(" m")
        self.bridge_radius.setVisible(False)  # Hidden by default
        mesh_form.addRow(self.radius_label, self.bridge_radius)
        self.radius_label.setVisible(False)
        self.bridge_radius.setVisible(False)

        # Create and hide multi span inputs:
        self.multi_span_dist_list_label = QLabel("Multi Span \nList")
        self.multi_span_dist_list = QLineEdit()
        self.multi_span_dist_list.setPlaceholderText("[span1, span2, span3,..]")  # Hint text
        self.multi_span_dist_list.setMaxLength(100) 
        self.nl_multi_label = QLabel("List of transv. \nmembers in \nevery span")
        self.nl_multi = QLineEdit()
        self.nl_multi.setPlaceholderText("[transv. members in span1, span2, span3,..]")  # Hint text
        self.nl_multi.setMaxLength(100)
        self.continuous = QCheckBox("Continuous spans", checked=True)
        self.stich_slab_x_spacing_label = QLabel("Spacing in spans")
        self.stich_slab_x_spacing = QDoubleSpinBox()
        self.stich_slab_x_spacing.setRange(0, 10)  # Adjust range as needed
        self.stich_slab_x_spacing.setValue(0.5)
        self.stich_slab_x_spacing.setSuffix(" m")
        
        mesh_form.addRow(self.multi_span_dist_list_label, self.multi_span_dist_list)
        mesh_form.addRow(self.nl_multi_label, self.nl_multi)
        mesh_form.addRow(self.continuous)
        mesh_form.addRow(self.stich_slab_x_spacing_label, self.stich_slab_x_spacing)
        self.stich_slab_x_spacing_label.setVisible(False)
        self.multi_span_dist_list.setVisible(False)
        self.multi_span_dist_list_label.setVisible(False)
        self.nl_multi.setVisible(False)
        self.nl_multi_label.setVisible(False)
        self.continuous.setVisible(False)
        self.stich_slab_x_spacing.setVisible(False)
        
        # Connect bridge type change to show/hide radius
        self.bridge_type.currentTextChanged.connect(self.toggle_radius_visibility)
        self.bridge_type.currentTextChanged.connect(self.toggle_multi_span_visibility)
        
        layout.addWidget(basic_group)
        layout.addWidget(mesh_group)
        layout.addStretch()
        tab.setLayout(layout)
        self.tabs.addTab(tab, "Geometry")

    def toggle_radius_visibility(self, bridge_type):
        """Show radius input only for curved bridges"""
        show = self.bridge_type.currentText() == "Curved"
        self.radius_label.setVisible(show)
        self.bridge_radius.setVisible(show)

    def toggle_multi_span_visibility(self, bridge_type):
        show = self.bridge_type.currentText() == "Multi-Span Straight"
        self.multi_span_dist_list_label.setVisible(show)
        self.multi_span_dist_list.setVisible(show)
        self.nl_multi.setVisible(show)
        self.continuous.setVisible(show)
        self.stich_slab_x_spacing.setVisible(show)
        self.nl_multi_label.setVisible(show)
        self.multi_span_dist_list.setVisible(show)
        self.stich_slab_x_spacing_label.setVisible(show)
        
    
    def create_materials_tab(self):
        """Material properties tab with nested group boxes"""
        tab = QWidget()
        main_layout = QVBoxLayout()
        
        # Outer group box - Material Properties
        outer_group = QGroupBox("Material Properties")
        outer_layout = QVBoxLayout()
        
        # Material Type selection
        self.material_type = QComboBox()
        self.material_type.addItems(["Concrete", "Steel"])
        self.material_type.currentTextChanged.connect(self.update_material_fields)
        
        # Inner group box - Grade Options
        inner_group = QGroupBox("Standard Code Options")
        inner_layout = QFormLayout()

        # Grade of material
        self.grade_box = QComboBox()
        self.update_grade_fields("AS5100-2017","Concrete") # Initialize with concrete options
        
        # Code of Material
        self.material_fc = QComboBox()
        self.update_material_fields("Concrete")  # Initialize with concrete options
        self.material_fc.currentTextChanged.connect(
    lambda: self.update_grade_fields(self.material_fc.currentText(), self.material_type.currentText())
)
        
        
        # Use Preset radio buttons
        self.use_preset_yes = QRadioButton("Use Preset")
        self.use_preset_no = QRadioButton("Custom Values")
        self.use_preset_yes.setChecked(True)
        
        # Custom material properties (hidden by default)
        self.custom_props_group = QGroupBox("Custom Material Properties")
        custom_props_layout = QFormLayout()
        
        self.material_ec = QDoubleSpinBox()
        self.material_ec.setRange(10000, 500000)
        self.material_ec.setValue(30000)
        self.material_ec.setSuffix(" MPa")
        
        self.material_density = QDoubleSpinBox()
        self.material_density.setRange(1000, 10000)
        self.material_density.setValue(2400)
        self.material_density.setSuffix(" kg/m³")
        
        self.material_poisson = QDoubleSpinBox()
        self.material_poisson.setRange(0.1, 0.5)
        self.material_poisson.setValue(0.2)
        self.material_poisson.setSingleStep(0.01)
        
        # Add widgets to inner layout
        inner_layout.addRow("Standard Code (Preset):", self.material_fc)
        inner_layout.addRow("Grade of material:", self.grade_box)
        inner_layout.addRow("Use Preset?:", self.use_preset_yes)
        inner_layout.addRow("", self.use_preset_no)
        
        # Add widgets to custom properties layout
        custom_props_layout.addRow("Elastic Modulus (E):", self.material_ec)
        custom_props_layout.addRow("Density:", self.material_density)
        custom_props_layout.addRow("Poisson's Ratio:", self.material_poisson)
        self.custom_props_group.setLayout(custom_props_layout)
        self.custom_props_group.setVisible(False)  # Hidden by default
        
        # Set inner group layout
        inner_group.setLayout(inner_layout)
        
        # Add widgets to outer layout
        outer_layout.addWidget(QLabel("Material Type:"))
        outer_layout.addWidget(self.material_type)
        outer_layout.addWidget(inner_group)
        outer_layout.addWidget(self.custom_props_group)
        
        # Set outer group layout
        outer_group.setLayout(outer_layout)
        
        # Add outer group to main layout
        main_layout.addWidget(outer_group)
        main_layout.addStretch()
        
        # Connect radio buttons to toggle custom properties
        self.use_preset_no.toggled.connect(
            lambda: self.custom_props_group.setVisible(self.use_preset_no.isChecked())
        )
        
        tab.setLayout(main_layout)
        self.tabs.addTab(tab, "Materials")
    
    def update_material_fields(self, material):
        """Update the compressive strength options based on material type"""
        self.material_fc.clear()
            
        if material == "Concrete":
            # AS5100-2017 concrete strength options
            self.material_fc.addItems(["AS5100-2017", "AASHTO-LRFD-8th"])
        elif material == "Steel":
            # Steel strength options
            self.material_fc.addItems([
                "AS5100.6-2004", "AASHTO-LRFD-8th"])
            
    def update_grade_fields(self, code, material):
        """Update the grade options based on code selected"""
        self.grade_box.clear()
            
        if code == "AS5100-2017":
            # AS5100-2017 concrete strength options
            self.grade_box.addItems(["32MPa", "40MPa", "50MPa",
                                       "65MPa", "80MPa",
                                       "100MPa"])
        if material == "Concrete":
            if code == "AASHTO-LRFD-8th":
                # AASHTO-LRFD-8th concrete strength options
                self.grade_box.addItems(["2.4ksi",
                                           "3.0ksi", "3.6ksi",
                                           "4.0ksi", "5.0ksi",
                                           "6.0ksi", "7.5ksi",
                                           "10.0ksi", "15.0ksi"])
        if material == "Steel":
            if code == "AASHTO-LRFD-8th":
                # AASHTO-LRFD-8th steel strength options
                self.grade_box.addItems(["A615-40",
                                        "A615-60",
                                        "A615-75",
                                        "A615-80",
                                        "A615-100",
                                        "A615-100"])
        if code == "AS5100.6-2004":
            # Steel strength options
            self.grade_box.addItems([
                "R250N",
                "D500N",
                "D500L"
            ])
    
    def create_sections_tab(self):
        """Cross-section input tab with specific properties"""
        tab = QWidget()
        main_layout = QVBoxLayout(tab)  # Set layout directly on the tab widget
        
        # Create a container widget for the scroll area
        container = QWidget()
        layout = QVBoxLayout(container)
        
        # Edge Longitudinal Section
        edge_long_group = QGroupBox("Edge Longitudinal Section")
        edge_long_form = QFormLayout(edge_long_group)  # Set layout directly on group
        
        self.edge_long_A = QDoubleSpinBox()
        self.edge_long_A.setRange(0.1, 10)
        self.edge_long_A.setValue(0.934)
        self.edge_long_A.setSingleStep(0.01)
        self.edge_long_A.setSuffix(" m²")
        
        self.edge_long_J = QDoubleSpinBox()
        self.edge_long_J.setRange(0.001, 1)
        self.edge_long_J.setValue(0.1857)
        self.edge_long_J.setSingleStep(0.01)
        self.edge_long_J.setSuffix(" m³")
        
        self.edge_long_Iz = QDoubleSpinBox()
        self.edge_long_Iz.setRange(0.001, 1)
        self.edge_long_Iz.setValue(0.3478)
        self.edge_long_Iz.setSingleStep(0.01)
        self.edge_long_Iz.setSuffix(" m⁴")
        
        self.edge_long_Iy = QDoubleSpinBox()
        self.edge_long_Iy.setRange(0.001, 1)
        self.edge_long_Iy.setValue(0.213602)
        self.edge_long_Iy.setSuffix(" m⁴")
        
        self.edge_long_Az = QDoubleSpinBox()
        self.edge_long_Az.setRange(0.001, 1)
        self.edge_long_Az.setValue(0.444795)
        self.edge_long_Az.setSuffix(" m²")
        
        self.edge_long_Ay = QDoubleSpinBox()
        self.edge_long_Ay.setRange(0.001, 1)
        self.edge_long_Ay.setValue(0.258704)
        self.edge_long_Ay.setSuffix(" m²")
        
        edge_long_form.addRow("Area (A)", self.edge_long_A)
        edge_long_form.addRow("Torsional Constant (J)", self.edge_long_J)
        edge_long_form.addRow("Moment of Inertia (Iz)", self.edge_long_Iz)
        edge_long_form.addRow("Moment of Inertia (Iy)", self.edge_long_Iy)
        edge_long_form.addRow("Shear Area (Az)", self.edge_long_Az)
        edge_long_form.addRow("Shear Area (Ay)", self.edge_long_Ay)
        
        # Longitudinal Section
        long_group = QGroupBox("Longitudinal Section")
        long_form = QFormLayout()
        
        self.long_A = QDoubleSpinBox()
        self.long_A.setRange(0.1, 10)
        self.long_A.setValue(1.025)
        self.long_A.setSuffix(" m²")
        
        self.long_J = QDoubleSpinBox()
        self.long_J.setRange(0.001, 1)
        self.long_J.setValue(0.1878)
        self.long_J.setSuffix(" m³")
        
        self.long_Iz = QDoubleSpinBox()
        self.long_Iz.setRange(0.001, 1)
        self.long_Iz.setValue(0.3694)
        self.long_Iz.setSuffix(" m⁴")
        
        self.long_Iy = QDoubleSpinBox()
        self.long_Iy.setRange(0.001, 1)
        self.long_Iy.setValue(0.3634)
        self.long_Iy.setSuffix(" m⁴")
        
        self.long_Az = QDoubleSpinBox()
        self.long_Az.setRange(0.001, 1)
        self.long_Az.setValue(0.4979)
        self.long_Az.setSuffix(" m²")
        
        self.long_Ay = QDoubleSpinBox()
        self.long_Ay.setRange(0.001, 1)
        self.long_Ay.setValue(0.309)
        self.long_Ay.setSuffix(" m²")
        
        long_form.addRow("Area (A)", self.long_A)
        long_form.addRow("Torsional Constant (J)", self.long_J)
        long_form.addRow("Moment of Inertia (Iz)", self.long_Iz)
        long_form.addRow("Moment of Inertia (Iy)", self.long_Iy)
        long_form.addRow("Shear Area (Az)", self.long_Az)
        long_form.addRow("Shear Area (Ay)", self.long_Ay)
        long_group.setLayout(long_form)
        
        # Transverse Section
        trans_group = QGroupBox("Transverse Section")
        trans_form = QFormLayout()
        
        self.trans_A = QDoubleSpinBox()
        self.trans_A.setRange(0.1, 10)
        self.trans_A.setValue(0.504)
        self.trans_A.setSuffix(" m²")
        
        self.trans_J = QDoubleSpinBox()
        self.trans_J.setRange(0.001, 1)
        self.trans_J.setValue(5.22303e-3)
        self.trans_J.setSuffix(" m³")
        
        self.trans_Iy = QDoubleSpinBox()
        self.trans_Iy.setRange(0.001, 1)
        self.trans_Iy.setValue(0.32928)
        self.trans_Iy.setSuffix(" m⁴")
        
        self.trans_Iz = QDoubleSpinBox()
        self.trans_Iz.setRange(0.001, 1)
        self.trans_Iz.setValue(1.3608e-3)
        self.trans_Iz.setSuffix(" m⁴")
        
        self.trans_Ay = QDoubleSpinBox()
        self.trans_Ay.setRange(0.001, 1)
        self.trans_Ay.setValue(0.42)
        self.trans_Ay.setSuffix(" m²")
        
        self.trans_Az = QDoubleSpinBox()
        self.trans_Az.setRange(0.001, 1)
        self.trans_Az.setValue(0.42)
        self.trans_Az.setSuffix(" m²")
        
        self.trans_unit_width = QCheckBox()
        self.trans_unit_width.setChecked(True)
        
        trans_form.addRow("Area (A)", self.trans_A)
        trans_form.addRow("Torsional Constant (J)", self.trans_J)
        trans_form.addRow("Moment of Inertia (Iy)", self.trans_Iy)
        trans_form.addRow("Moment of Inertia (Iz)", self.trans_Iz)
        trans_form.addRow("Shear Area (Ay)", self.trans_Ay)
        trans_form.addRow("Shear Area (Az)", self.trans_Az)
        trans_form.addRow("Unit Width", self.trans_unit_width)
        trans_group.setLayout(trans_form)
        
        # End Transverse Section
        end_trans_group = QGroupBox("End Transverse Section")
        end_trans_form = QFormLayout()
        
        self.end_trans_A = QDoubleSpinBox()
        self.end_trans_A.setRange(0.1, 10)
        self.end_trans_A.setValue(0.252)
        self.end_trans_A.setSuffix(" m²")
        
        self.end_trans_J = QDoubleSpinBox()
        self.end_trans_J.setRange(0.001, 1)
        self.end_trans_J.setValue(2.5012e-3)
        self.end_trans_J.setSuffix(" m³")
        
        self.end_trans_Iy = QDoubleSpinBox()
        self.end_trans_Iy.setRange(0.001, 1)
        self.end_trans_Iy.setValue(0.04116)
        self.end_trans_Iy.setSuffix(" m⁴")
        
        self.end_trans_Iz = QDoubleSpinBox()
        self.end_trans_Iz.setRange(0.001, 1)
        self.end_trans_Iz.setValue(0.6804e-3)
        self.end_trans_Iz.setSuffix(" m⁴")
        
        self.end_trans_Ay = QDoubleSpinBox()
        self.end_trans_Ay.setRange(0.001, 1)
        self.end_trans_Ay.setValue(0.21)
        self.end_trans_Ay.setSuffix(" m²")
        
        self.end_trans_Az = QDoubleSpinBox()
        self.end_trans_Az.setRange(0.001, 1)
        self.end_trans_Az.setValue(0.21)
        self.end_trans_Az.setSuffix(" m²")
        
        end_trans_form.addRow("Area (A)", self.end_trans_A)
        end_trans_form.addRow("Torsional Constant (J)", self.end_trans_J)
        end_trans_form.addRow("Moment of Inertia (Iy)", self.end_trans_Iy)
        end_trans_form.addRow("Moment of Inertia (Iz)", self.end_trans_Iz)
        end_trans_form.addRow("Shear Area (Ay)", self.end_trans_Ay)
        end_trans_form.addRow("Shear Area (Az)", self.end_trans_Az)
        end_trans_group.setLayout(end_trans_form)
        
        # Add all section groups to the scrollable layout
        layout.addWidget(edge_long_group)
        layout.addWidget(long_group)
        layout.addWidget(trans_group)
        layout.addWidget(end_trans_group)
        layout.addStretch()
        
        # Set up scroll area
        scroll = QScrollArea()
        scroll.setWidgetResizable(True)
        scroll.setWidget(container)
        
        # Add scroll area to main layout
        main_layout.addWidget(scroll)
        
        self.tabs.addTab(tab, "Sections")
    
    def create_members_tab(self):
        """Member assignment tab"""
        tab = QWidget()
        layout = QVBoxLayout()
        
        # Interior Beams Group
        int_beam_group = QGroupBox("External to Internal distance")
        int_beam_form = QFormLayout()
        
        self.int_beam_spacing = QDoubleSpinBox()
        self.int_beam_spacing.setRange(0.1, 10)
        self.int_beam_spacing.setValue(2.5)
        self.int_beam_spacing.setSingleStep(0.1)
        self.int_beam_spacing.setSuffix(" m")
        
        int_beam_form.addRow("Spacing", self.int_beam_spacing)
        int_beam_group.setLayout(int_beam_form)
        
        # Edge Beams Group
        edge_beam_group = QGroupBox("Edge Beams")
        edge_beam_form = QFormLayout()
        
        self.edge_beam_offset = QDoubleSpinBox()
        self.edge_beam_offset.setRange(0.1, 5)
        self.edge_beam_offset.setValue(0.5)
        self.edge_beam_offset.setSingleStep(0.1)
        self.edge_beam_offset.setSuffix(" m")
        
        edge_beam_form.addRow("Edge Offset", self.edge_beam_offset)
        edge_beam_group.setLayout(edge_beam_form)
        
        layout.addWidget(int_beam_group)
        layout.addWidget(edge_beam_group)
        layout.addStretch()
        tab.setLayout(layout)
        self.tabs.addTab(tab, "Members")
    
    def create_loads_tab(self):
        """Load cases tab"""
        tab = QWidget()
        layout = QVBoxLayout()
        
        # Dead Load Group
        dl_group = QGroupBox("Dead Load")
        dl_form = QFormLayout()
        
        self.dl_magnitude = QDoubleSpinBox()
        self.dl_magnitude.setRange(0, 100)
        self.dl_magnitude.setValue(5)
        self.dl_magnitude.setSuffix(" kN/m²")
        
        self.dl_direction = QComboBox()
        self.dl_direction.addItems(["-Y (Downward)", "+Y (Upward)", "-Z", "+Z"])
        
        dl_form.addRow("Magnitude", self.dl_magnitude)
        dl_form.addRow("Direction", self.dl_direction)
        dl_group.setLayout(dl_form)
        
        # Live Load Group
        ll_group = QGroupBox("Live Load")
        ll_form = QFormLayout()
        
        self.ll_magnitude = QDoubleSpinBox()
        self.ll_magnitude.setRange(0, 100)
        self.ll_magnitude.setValue(10)
        self.ll_magnitude.setSuffix(" kN/m²")
        
        self.ll_direction = QComboBox()
        self.ll_direction.addItems(["-Y (Downward)", "+Y (Upward)", "-Z", "+Z"])
        
        ll_form.addRow("Magnitude", self.ll_magnitude)
        ll_form.addRow("Direction", self.ll_direction)
        ll_group.setLayout(ll_form)
        
        layout.addWidget(dl_group)
        layout.addWidget(ll_group)
        layout.addStretch()
        tab.setLayout(layout)
        self.tabs.addTab(tab, "Loads")
    
    def create_analysis_tab(self):
        """Analysis settings tab (simplified)"""
        tab = QWidget()
        layout = QVBoxLayout()
        
        # Output Settings Group
        output_group = QGroupBox("Output Settings")
        output_form = QFormLayout()
        
        self.output_displacements = QComboBox()
        self.output_displacements.addItems(["Yes", "No"])
        self.output_displacements.setCurrentIndex(0)
        
        self.output_forces = QComboBox()
        self.output_forces.addItems(["Yes", "No"])
        self.output_forces.setCurrentIndex(0)
        
        self.output_stresses = QComboBox()
        self.output_stresses.addItems(["Yes", "No"])
        
        output_form.addRow("Output Displacements", self.output_displacements)
        output_form.addRow("Output Forces", self.output_forces)
        output_form.addRow("Output Stresses", self.output_stresses)
        output_group.setLayout(output_form)
        
        layout.addWidget(output_group)
        layout.addStretch()
        tab.setLayout(layout)
        self.tabs.addTab(tab, "Analysis")

class BridgeAnalysisGUI(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Bridge Geometry Studio")  # Updated name
        #self.setWindowIcon(QIcon("ospgrillage_logo.png"))  # Add your icon file
        # Add this stylesheet
        self.setStyleSheet("""
            QMainWindow {
                background-color: #f0f0f0;
            }
            QGroupBox {
                border: 1px solid #cccccc;
                border-radius: 4px;
                margin-top: 1ex;
                font-weight: bold;
            }
            QGroupBox::title {
                subcontrol-origin: margin;
                left: 10px;
                padding: 0 3px;
            }
            QTabWidget::pane {
                border: 1px solid #cccccc;
                top: -1px;
            }
            QTabBar::tab {
                background: #e0e0e0;
                border: 1px solid #cccccc;
                padding: 8px;
                min-width: 100px;
            }
            QTabBar::tab:selected {
                background: #ffffff;
                border-bottom-color: #ffffff;
            }
            QLineEdit, QComboBox, QSpinBox, QDoubleSpinBox, QTextEdit {
                border: 1px solid #cccccc;
                border-radius: 3px;
                padding: 3px;
                min-height: 20px;
            }
            QPushButton {
                background-color: #0078d4;
                color: white;
                border: none;
                padding: 5px 15px;
                border-radius: 4px;
            }
            QPushButton:hover {
                background-color: #006cbd;
            }
        """)
        self.setWindowTitle("ospgui")
        self.setGeometry(100, 100, 1400, 900)
        
        # Initialize variables to store user inputs
        self.bridge_params = {}
        self.generated_code = ""
        
        # Create UI components
        self.create_menu_bar()
        self.create_tool_bar()
        self.create_status_bar()
        self.create_main_content()

    def open_file(self):
        options = QFileDialog.Options()
        file_name, _ = QFileDialog.getOpenFileName(
            self,
            "Open Python Script",
            "",
            "Python Files (*.py);;All Files (*)",
            options=options
        )
        
        if file_name:
            try:
                with open(file_name, 'r') as f:
                    content = f.read()
                    self.code_tab.setPlainText(content)
                    self.statusbar.showMessage(f"Loaded: {os.path.basename(file_name)}", 3000)
            except Exception as e:
                QMessageBox.critical(self, "Error", f"Could not open file:\n{str(e)}")

    def show_about(self):
        QMessageBox.about(self, "About This Application", 
                          "This is a simple GUI Application for accesing Ospgrillage geometry features with ease.\nVersion 1.0\nMay 2025")
         
    def create_menu_bar(self):
        """Create the main menu bar"""
        menubar = self.menuBar()
        
        # File menu
        file_menu = menubar.addMenu("File")
        
        new_action = QAction(QIcon.fromTheme("document-new"), "New", self)
        new_action.setShortcut("Ctrl+N")
        file_menu.addAction(new_action)
        
        open_action = QAction(QIcon.fromTheme("document-open"), "Open", self)
        open_action.setShortcut("Ctrl+O")
        open_action.triggered.connect(self.open_file)
        file_menu.addAction(open_action)
        
        save_action = QAction(QIcon.fromTheme("document-save"), "Save", self)
        save_action.setShortcut("Ctrl+S")
        save_action.triggered.connect(self.save_code)
        file_menu.addAction(save_action)
        
        file_menu.addSeparator()
        
        exit_action = QAction(QIcon.fromTheme("application-exit"), "Exit", self)
        exit_action.setShortcut("Ctrl+Q")
        exit_action.triggered.connect(self.close)
        file_menu.addAction(exit_action)
        
##        # Tools menu
##        tools_menu = menubar.addMenu("Tools")
##        
##        settings_action = QAction("Settings", self)
##        tools_menu.addAction(settings_action)
        
        # Help menu
        help_menu = menubar.addMenu("Help")
        
        about_action = QAction("About", self)
        help_menu.addAction(about_action)
        about_action.triggered.connect(self.show_about)

    
    def create_tool_bar(self):
        """Create the main toolbar"""
        toolbar = QToolBar("Main Toolbar")
        self.addToolBar(toolbar)
        
        # Run Analysis action
        self.run_action = QAction(QIcon.fromTheme("media-playback-start"), "Create Geometry", self)
        self.run_action.setShortcut("F5")
        self.run_action.triggered.connect(self.run_analysis)
        toolbar.addAction(self.run_action)
        
        # Separator
        toolbar.addSeparator()
        
        # Zoom tools
##        zoom_in = QAction(QIcon.fromTheme("zoom-in"), "Zoom In", self)
##        zoom_out = QAction(QIcon.fromTheme("zoom-out"), "Zoom Out", self)
##        toolbar.addAction(zoom_in)
##        toolbar.addAction(zoom_out)
    
    def create_status_bar(self):
        """Create the status bar"""
        self.statusbar = QStatusBar()
        self.setStatusBar(self.statusbar)
        self.statusbar.showMessage("Ready")
    
    def create_main_content(self):
        """Create the main content area with tabs"""
        central_widget = QWidget()
        self.setCentralWidget(central_widget)
        
        # Main layout
        main_layout = QHBoxLayout()
        central_widget.setLayout(main_layout)
        
        # Left panel - Input parameters
        self.input_panel = BridgeInputWidget()
        self.input_panel.btn_apply.clicked.connect(self.apply_changes)
        self.input_panel.btn_run.clicked.connect(self.run_analysis)
        main_layout.addWidget(self.input_panel, stretch=1)
        
        # Right panel - Visualization and code tabs
        self.right_panel = QTabWidget()
        
        # Visualization tab
##        self.viz_tab = QLabel("3D Visualization will appear here")
##        self.viz_tab.setAlignment(Qt.AlignCenter)
        
        # Code view tab
        self.code_tab = QTextEdit()
        self.code_tab.setStyleSheet("font-family: monospace; font-size: 10pt;")
        self.code_tab.setLineWrapMode(QTextEdit.NoWrap)
        
        # Add tabs to right panel
##        self.right_panel.addTab(self.viz_tab, "Visualization")
        self.right_panel.addTab(self.code_tab, "Code View")
        
        # Add panels to main layout
        main_layout.addWidget(self.input_panel, stretch=1)
        main_layout.addWidget(self.right_panel, stretch=2)
    
    def apply_changes(self):
        """Handle Apply Changes button click"""
        try:
            # Collect all parameters from input fields
            self.collect_parameters()
            
            # Generate the ospgrillage code
            self.generate_ospgrillage_code()
            
            # Update code view
            self.code_tab.setPlainText(self.generated_code)
            
            # Update status bar
            self.statusbar.showMessage("Parameters applied successfully", 3000)
            
        except Exception as e:
            self.statusbar.showMessage(f"Error applying changes: {(e)}", 5000)
            print(f"Error in apply_changes: {str(e)}")
            
    def collect_parameters(self):
        """Collect all parameters from input fields"""
        self.bridge_params = {
            'geometry': {
                'name': self.input_panel.bridge_name.text(),
                'length': self.input_panel.bridge_length.value(),
                'width': self.input_panel.bridge_width.value(),
                'left_skew_angle': self.input_panel.left_skew_angle.value(),
                'right_skew_angle': self.input_panel.right_skew_angle.value(),
                #'num_spans': self.input_panel.num_spans.value(),
                'long_beams': self.input_panel.long_beams.value(),
                'trans_beams': self.input_panel.trans_beams.value(),
                'mesh_type': self.input_panel.mesh_type.currentText(),
                'need_command_file': self.input_panel.opensees_file.isChecked(),
                'bridge_type': self.input_panel.bridge_type.currentText(),
                'radius': self.input_panel.bridge_radius.value(),
                'multi_span_dist_list': self.input_panel.multi_span_dist_list.text(),
                'nl_multi': self.input_panel.nl_multi.text(),
                'continuous': self.input_panel.continuous.isChecked(),
                'stich_slab_x_spacing': self.input_panel.stich_slab_x_spacing.value()
                
            
            },
            'materials': {
                'type': self.input_panel.material_type.currentText(),
                'ec': self.input_panel.material_ec.value(),
                'density': self.input_panel.material_density.value(),
                'poisson': self.input_panel.material_poisson.value(),
                'preset_yes': self.input_panel.use_preset_yes.isChecked(),
                'preset_no': self.input_panel.use_preset_no.isChecked(),
                'code': self.input_panel.material_fc.currentText(),
                'grade': self.input_panel.grade_box.currentText()
            },
            'sections': {
                'edge_longitudinal': {
                    'A': self.input_panel.edge_long_A.value(),
                    'J': self.input_panel.edge_long_J.value(),
                    'Iz': self.input_panel.edge_long_Iz.value(),
                    'Iy': self.input_panel.edge_long_Iy.value(),
                    'Az': self.input_panel.edge_long_Az.value(),
                    'Ay': self.input_panel.edge_long_Ay.value()
                },
                'longitudinal': {
                    'A': self.input_panel.long_A.value(),
                    'J': self.input_panel.long_J.value(),
                    'Iz': self.input_panel.long_Iz.value(),
                    'Iy': self.input_panel.long_Iy.value(),
                    'Az': self.input_panel.long_Az.value(),
                    'Ay': self.input_panel.long_Ay.value()
                },
                'transverse': {
                    'A': self.input_panel.trans_A.value(),
                    'J': self.input_panel.trans_J.value(),
                    'Iy': self.input_panel.trans_Iy.value(),
                    'Iz': self.input_panel.trans_Iz.value(),
                    'Ay': self.input_panel.trans_Ay.value(),
                    'Az': self.input_panel.trans_Az.value(),
                    'unit_width': self.input_panel.trans_unit_width.isChecked()
                },
                'end_transverse': {
                    'A': self.input_panel.end_trans_A.value(),
                    'J': self.input_panel.end_trans_J.value(),
                    'Iy': self.input_panel.end_trans_Iy.value(),
                    'Iz': self.input_panel.end_trans_Iz.value(),
                    'Ay': self.input_panel.end_trans_Ay.value(),
                    'Az': self.input_panel.end_trans_Az.value()
                }
            },
##            'loads': {
##                'dead_load': {
##                    'magnitude': self.input_panel.dl_magnitude.value(),
##                    'direction': self.input_panel.dl_direction.currentText()
##                },
##                'live_load': {
##                    'magnitude': self.input_panel.ll_magnitude.value(),
##                    'direction': self.input_panel.ll_direction.currentText()
##                }
##            },
##            'analysis': {
##                'output_displacements': self.input_panel.output_displacements.currentText(),
##                'output_forces': self.input_panel.output_forces.currentText(),
##                'output_stresses': self.input_panel.output_stresses.currentText()
##            }
        }
    
    def generate_ospgrillage_code(self):
        if self.bridge_params['geometry']['need_command_file'] == True:
            plot_or_save = f"""model.create_osp_model(pyfile=True)"""
        else:
            plot_or_save = f"""model.create_osp_model(pyfile=False)
og.opsv.plot_model(az_el=(-90, 0),element_labels=0, node_labels=0)
fig = og.plt.gcf()
fig.set_size_inches(8, 8)
og.plt.show()"""

        ##Define material(Preset or custom)
        if self.bridge_params['materials']['preset_yes'] == False:
            material_code = f"""material = og.create_material(
    material='{self.bridge_params['materials']['type'].lower()}',
    E={self.bridge_params['materials']['ec']}*GPa,
    v= {self.bridge_params['materials']['poisson']},
    rho={self.bridge_params['materials']['density']}*kN/m3 
)"""
        else:
            material_code = f"""material = og.create_material(material="{self.bridge_params['materials']['type'].lower()}",
code="{self.bridge_params['materials']['code']}", grade="{self.bridge_params['materials']['grade']}")"""
        """Generate ospgrillage Python code from collected parameters"""

        if self.bridge_params['geometry']['bridge_type'] == "Curved":
            create_model = f"""model = og.create_grillage(
    bridge_name="{self.bridge_params['geometry']['name']}",
    long_dim={self.bridge_params['geometry']['length']} * m,
    width={self.bridge_params['geometry']['width']} * m,
    skew=[{self.bridge_params['geometry']['left_skew_angle']},{self.bridge_params['geometry']['right_skew_angle']}],
    num_long_grid={self.bridge_params['geometry']['long_beams']},  # Number of grid lines
    num_trans_grid={self.bridge_params['geometry']['trans_beams']},
    mesh_type="{self.bridge_params['geometry']['mesh_type']}",  # ('Ortho' or 'Oblique')
    mesh_radius={self.bridge_params['geometry']['radius']},
)"""
        if self.bridge_params['geometry']['bridge_type'] == "Straight":
            create_model = f"""# Create grillage model
model = og.create_grillage(
    bridge_name="{self.bridge_params['geometry']['name']}",
    long_dim={self.bridge_params['geometry']['length']} * m,
    width={self.bridge_params['geometry']['width']} * m,
    skew=[{self.bridge_params['geometry']['left_skew_angle']},{self.bridge_params['geometry']['right_skew_angle']}],
    num_long_grid={self.bridge_params['geometry']['long_beams']},  # Number of grid lines
    num_trans_grid={self.bridge_params['geometry']['trans_beams']},
    ext_to_int_dist = {self.input_panel.int_beam_spacing.value()} * m,
    edge_beam_dist={self.input_panel.edge_beam_offset.value()} * m,
    mesh_type="{self.bridge_params['geometry']['mesh_type']}"  # ('Ortho' or 'Oblique')
)"""
        if self.bridge_params['geometry']['bridge_type'] == "Multi-Span Straight":
            create_model =f"""model = og.create_grillage(
        bridge_name="{self.bridge_params['geometry']['name']}",
        long_dim={self.bridge_params['geometry']['length']} * m,
        width={self.bridge_params['geometry']['width']} * m,
        skew=[{self.bridge_params['geometry']['left_skew_angle']},{self.bridge_params['geometry']['right_skew_angle']}],
        num_long_grid={self.bridge_params['geometry']['long_beams']},  # Number of grid lines
        num_trans_grid={self.bridge_params['geometry']['trans_beams']},
        edge_beam_dist={self.input_panel.edge_beam_offset.value()} * m,
        mesh_type="{self.bridge_params['geometry']['mesh_type']}",  # ('Ortho' or 'Oblique')
        multi_span_dist_list={self.bridge_params['geometry']['multi_span_dist_list']},
        multi_span_num_points={self.bridge_params['geometry']['nl_multi']},
        continuous={self.bridge_params['geometry']['continuous']},
        non_cont_spacing_x={self.bridge_params['geometry']['stich_slab_x_spacing']},
    )"""
        # Unit definitions
        units_code = """# Unit definitions
kilo = 1e3
milli = 1e-3
N = 1
m = 1
mm = milli * m
m2 = m ** 2
m3 = m ** 3
m4 = m ** 4
kN = kilo * N
MPa = N / ((mm) ** 2)
GPa = kilo * MPa
"""
        
        # Material definition
        material_code = f"""
# Material definition
{material_code}
"""
        
        # Section definitions
        sections_code = f"""
# Section definitions
edge_longitudinal_section = og.create_section(
    A={self.bridge_params['sections']['edge_longitudinal']['A']} * m2,
    J={self.bridge_params['sections']['edge_longitudinal']['J']} * m3,
    Iz={self.bridge_params['sections']['edge_longitudinal']['Iz']} * m4,
    Iy={self.bridge_params['sections']['edge_longitudinal']['Iy']} * m4,
    Az={self.bridge_params['sections']['edge_longitudinal']['Az']} * m2,
    Ay={self.bridge_params['sections']['edge_longitudinal']['Ay']} * m2
)

longitudinal_section = og.create_section(
    A={self.bridge_params['sections']['longitudinal']['A']} * m2,
    J={self.bridge_params['sections']['longitudinal']['J']} * m3,
    Iz={self.bridge_params['sections']['longitudinal']['Iz']} * m4,
    Iy={self.bridge_params['sections']['longitudinal']['Iy']} * m4,
    Az={self.bridge_params['sections']['longitudinal']['Az']} * m2,
    Ay={self.bridge_params['sections']['longitudinal']['Ay']} * m2
)

transverse_section = og.create_section(
    A={self.bridge_params['sections']['transverse']['A']} * m2,
    J={self.bridge_params['sections']['transverse']['J']} * m3,
    Iy={self.bridge_params['sections']['transverse']['Iy']} * m4,
    Iz={self.bridge_params['sections']['transverse']['Iz']} * m4,
    Ay={self.bridge_params['sections']['transverse']['Ay']} * m2,
    Az={self.bridge_params['sections']['transverse']['Az']} * m2,
    unit_width={str(self.bridge_params['sections']['transverse']['unit_width'])}
)


end_transverse_section = og.create_section(
    A={self.bridge_params['sections']['end_transverse']['A']} * m2,
    J={self.bridge_params['sections']['end_transverse']['J']} * m3,
    Iy={self.bridge_params['sections']['end_transverse']['Iy']} * m4,
    Iz={self.bridge_params['sections']['end_transverse']['Iz']} * m4,
    Ay={self.bridge_params['sections']['end_transverse']['Ay']} * m2,
    Az={self.bridge_params['sections']['end_transverse']['Az']} * m2
)

longitudinal_beam = og.create_member(section=longitudinal_section, material=material)

edge_longitudinal_beam = og.create_member(
    section=edge_longitudinal_section, material=material
)
transverse_slab = og.create_member(section=transverse_section, material=material)

end_transverse_slab = og.create_member(
    section=end_transverse_section, material=material
)


"""
        
        # Create grillage model
        model_code = f"""{create_model}
"""
        
        # Member assignments
        members_code = f"""
# Member assignments
model.set_member(longitudinal_beam, member="interior_main_beam")
model.set_member(longitudinal_beam, member="exterior_main_beam_1")
model.set_member(longitudinal_beam, member="exterior_main_beam_2")
model.set_member(edge_longitudinal_beam, member="edge_beam")
model.set_member(transverse_slab, member="transverse_slab")
model.set_member(end_transverse_slab, member="start_edge")
model.set_member(end_transverse_slab, member="end_edge")
"""

        visualization_code = f"""{plot_or_save}
"""
##        # Load cases
##        loads_code = f"""
### Load cases
##dead_load_case = og.load.create_load_case(name='DeadLoad')
##live_load_case = og.load.create_load_case(name='LiveLoad')
##
### Create load vertices (example for uniform load)
##load_area = og.load.create_area(
##    x1=0, z1=0,  # Start point
##    x2={self.bridge_params['geometry']['length']} * m, 
##    z2={self.bridge_params['geometry']['width']} * m,  # End point
##    p={self.bridge_params['loads']['dead_load']['magnitude']} * kN/m2
##)
##
### Add loads to load cases
##dead_load_case.add_load(
##    og.load.create_load(
##        name='DeadLoad',
##        load_type='Area',
##        points=[load_area],
##        direction='{self.bridge_params['loads']['dead_load']['direction'][0:2]}'  # Just the direction part
##    )
##)
##
##live_load_case.add_load(
##    og.load.create_load(
##        name='LiveLoad',
##        load_type='Area',
##        points=[load_area],
##        magnitude={self.bridge_params['loads']['live_load']['magnitude']} * kN/m2,
##        direction='{self.bridge_params['loads']['live_load']['direction'][0:2]}'
##    )
##)
##
### Add load cases to model
##model.add_load_case(dead_load_case)
##model.add_load_case(live_load_case)
##"""
##        
##        # Analysis and results
##        output_lines = []
##
##        if self.bridge_params['analysis']['output_displacements'] == 'Yes':
##            output_lines.append("displacements = results.get_displacements()")
##        if self.bridge_params['analysis']['output_forces'] == 'Yes':
##            output_lines.append("forces = results.get_forces()")
##        if self.bridge_params['analysis']['output_stresses'] == 'Yes':
##            output_lines.append("stresses = results.get_stresses()")
##
##        summary_lines = [
##            'print()',
##            'print("- Number of nodes: {}".format(len(results.node)))',
##            'print("- Number of elements: {}".format(len(results.element)))'
##        ]
##
##        if self.bridge_params['analysis']['output_displacements'] == 'Yes':
##            summary_lines.append('print("- Maximum displacement: {:.6f} m".format(displacements.max().values))')
##        if self.bridge_params['analysis']['output_forces'] == 'Yes':
##            summary_lines.append('print("- Maximum bending moment: {:.2f} kN·m".format(forces.sel(component=\'Mz\').max().values))')
##        if self.bridge_params['analysis']['output_stresses'] == 'Yes':
##            summary_lines.append('print("- Maximum stress: {:.2f} MPa".format(stresses.max().values))')
##
##        analysis_code = f"""
### Analysis and results
##model.analyze()
##
### Get results
##results = model.get_results()
##print("Analysis completed successfully!")
##
### Initialize variables for results
##displacements = None
##forces = None
##stresses = None
##
### Store requested results
##{chr(10).join(output_lines)}
##
### Print summary of results
##{chr(10).join(summary_lines)}
##"""



        
        # Combine all code sections
        self.generated_code = f"""# Bridge Geometry Script
# Generated by ospgui

import ospgrillage as og
from math import *

{units_code}
{material_code}
{sections_code}
{model_code}
{members_code}
{visualization_code}
"""
##{loads_code}
##{analysis_code}


    def save_code(self):
        """Save the generated code to a file with proper error handling"""
        if not hasattr(self, 'generated_code') or not self.generated_code:
            QMessageBox.warning(self, "Warning", "No code to save. Please generate code first.")
            return

        options = QFileDialog.Options()
        file_name, _ = QFileDialog.getSaveFileName(
            self,
            "Save Python Script",
            "",  # Start in current directory
            "Python Files (*.py);;All Files (*)",
            options=options
        )
        
        if not file_name:  # User cancelled the dialog
            return
        
        try:
            # Ensure .py extension
            if not file_name.lower().endswith('.py'):
                file_name += '.py'
            
            # Create parent directories if needed
            os.makedirs(os.path.dirname(file_name), exist_ok=True)
            
            # Write with proper encoding and line endings
            with open(file_name, 'w', encoding='utf-8') as f:
                f.write(self.generated_code)
            
            # Show success message
            self.statusbar.showMessage(f"Script successfully saved to {os.path.abspath(file_name)}", 5000)
            
        except PermissionError:
            QMessageBox.critical(self, "Error", 
                f"Permission denied. Could not save to:\n{file_name}")
        except OSError as e:
            QMessageBox.critical(self, "Error", 
                f"Failed to save file:\n{str(e)}")
        except Exception as e:
            QMessageBox.critical(self, "Error", 
                f"Unexpected error saving file:\n{str(e)}")

    def run_analysis(self):
        """Handle Run Analysis button click"""
        try:
            # First apply any changes
            #self.apply_changes()
            
            # Update status bar
            self.statusbar.showMessage("Running analysis...")
            
            # Get the current code (either generated or manually edited)
            current_code = self.code_tab.toPlainText()
            
            # Create a dictionary to capture the output
            output = {}
            
            # Execute the code in a separate namespace
            namespace = {'og': None, 'print': lambda x: output.setdefault('print', []).append(x)}
            
            # Try to import ospgrillage
            try:
                import ospgrillage as og
                namespace['og'] = og
            except ImportError:
                QMessageBox.critical(self, "Error", "ospgrillage package not found. Please install it first.")
                return
            
            # Execute the code
            try:
                exec(current_code, namespace)
                
                # Update visualization tab with results
                result_text = "Analysis completed successfully!\n\n"
                
                # Add printed output if any
                if 'print' in output:
                    result_text += "Output:\n" + "\n".join(output['print']) + "\n\n"
                
                # Add basic result information
                if 'results' in namespace:
                    results = namespace['results']
                    result_text += f"Results summary:\n"
                    result_text += f"- Number of nodes: {len(results.node)}\n"
                    result_text += f"- Number of elements: {len(results.element)}\n"
                    
                    if 'displacements' in namespace:
                        max_disp = namespace['displacements'].max().values
                        result_text += f"- Maximum displacement: {max_disp:.6f} m\n"
                    
                    if 'forces' in namespace:
                        max_moment = namespace['forces'].sel(component='Mz').max().values
                        result_text += f"- Maximum bending moment: {max_moment:.2f} kN·m"
                
##                self.viz_tab.setText(result_text)
##                self.statusbar.showMessage("Analysis completed successfully", 5000)
                
            except Exception as e:
                error_msg = f"Error during analysis:\n{str(e)}"
##                self.viz_tab.setText(error_msg)
                self.statusbar.showMessage("Analysis failed", 5000)
                QMessageBox.critical(self, "Analysis Error", error_msg)
            
        except Exception as e:
            self.statusbar.showMessage(f"Analysis failed: {str(e)}", 5000)
            QMessageBox.critical(self, "Error", f"Failed to create geometry: {str(e)}")

def main():
    app = QApplication(sys.argv)
    window = BridgeAnalysisGUI()
    window.show()
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



from ospgrillage import __version__ as version
from fixtures import *

sys.path.insert(0, os.path.abspath("../"))


def test_envelope(bridge_model_42_negative):
    # test functionality of Envelope class and output
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative

    # create reference line load
    p = 10000
    p2 = 20000
    p3 = 30000  # duplicate of 2nd but with different mag
    barrierpoint_1 = og.create_load_vertex(x=5, z=1, p=p)
    barrierpoint_2 = og.create_load_vertex(x=10, z=7, p=p)
    barrierpoint_3 = og.create_load_vertex(x=10, z=2, p=p2)
    barrierpoint_4 = og.create_load_vertex(x=5, z=5, p=p2)
    barrierpoint_5 = og.create_load_vertex(x=10, z=2, p=p3)
    barrierpoint_6 = og.create_load_vertex(x=5, z=5, p=p3)
    # add moving load case
    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(2, 0, 2, 50)
    )  # Single point load 50 N
    Barrier = og.create_load(
        loadtype="line",
        name="Barrier curb load",
        point1=barrierpoint_1,
        point2=barrierpoint_2,
    )
    Barrier2 = og.create_load(
        loadtype="line", name="Barrieload", point1=barrierpoint_3, point2=barrierpoint_4
    )
    Barrier3 = og.create_load(
        loadtype="line", name="Barrieload", point1=barrierpoint_5, point2=barrierpoint_6
    )
    Patch1 = og.create_load(
        loadtype="patch",
        point1=barrierpoint_1,
        point2=barrierpoint_3,
        point3=barrierpoint_2,
        point4=barrierpoint_4,
    )

    barrierpoint_1 = og.create_load_vertex(x=6, z=2, p=0)
    barrierpoint_2 = og.create_load_vertex(x=11, z=8, p=0)
    barrierpoint_3 = og.create_load_vertex(x=11, z=3, p=0)
    barrierpoint_4 = og.create_load_vertex(x=6, z=6, p=0)

    Patch2 = og.create_load(
        loadtype="patch",
        point1=barrierpoint_1,
        point2=barrierpoint_3,
        point3=barrierpoint_2,
        point4=barrierpoint_4,
    )
    # create basic load case
    barrier_load_case = og.create_load_case(name="Barrier")
    # barrier_load_case.add_load_groups(Barrier)  # ch
    barrier_load_case.add_load(Patch1)  # ch
    # 2nd
    barrier_load_case2 = og.create_load_case(name="Barrier2")
    # barrier_load_case2.add_load_groups(Barrier2)
    barrier_load_case2.add_load(Patch2)
    # 3rd
    # barrier_load_case3 = og.create_load_case(name="Barrier3")
    # barrier_load_case3.add_load_groups(Barrier3)

    # adding load cases to model
    example_bridge.add_load_case(barrier_load_case)
    example_bridge.add_load_case(barrier_load_case2)
    # example_bridge.add_load_case(barrier_load_case3)

    single_path = og.create_moving_path(
        start_point=og.Point(2, 0, 2), end_point=og.Point(4, 0, 3), increments=3
    )  # create path object
    move_point = og.create_moving_load(name="single_moving_point")
    move_point.set_path(single_path)
    move_point.add_load(load_obj=front_wheel)
    example_bridge.add_load_case(move_point)

    example_bridge.analyze()
    results = example_bridge.get_results()
    comb_results = example_bridge.get_results(
        combinations={"Barrier": 1, "single_moving_point": 2}
    )
    # maxY = results.sel(Component='dy').max()
    envelope = og.create_envelope(
        ds=comb_results, load_effect="dy", array="displacements"
    )
    max_disp = envelope.get()
    print(max_disp)
    move_point.query(
        incremental_lc_name="single_moving_point at global position [2.00,0.00,2.00]"
    )
    # print(comb_results)
    print(og.ops.nodeDisp(25)[1])


def test_plot_force(bridge_model_42_negative):
    # test functionality of plot_force and its output
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative
    # og.opsv.plot_model()
    # og.plt.show()
    # create reference line load
    p = 10000
    p2 = 20000
    p3 = 30000  # duplicate of 2nd but with different mag
    barrierpoint_1 = og.create_load_vertex(x=5, z=1, p=p)
    barrierpoint_2 = og.create_load_vertex(x=10, z=7, p=p)
    barrierpoint_3 = og.create_load_vertex(x=10, z=2, p=p2)
    barrierpoint_4 = og.create_load_vertex(x=5, z=5, p=p2)
    barrierpoint_5 = og.create_load_vertex(x=10, z=2, p=p3)
    barrierpoint_6 = og.create_load_vertex(x=5, z=5, p=p3)
    # add moving load case
    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(2, 0, 2, 50)
    )  # Single point load 50 N
    Barrier = og.create_load(
        loadtype="line",
        name="Barrier curb load",
        point1=barrierpoint_1,
        point2=barrierpoint_2,
    )
    Barrier2 = og.create_load(
        loadtype="line", name="Barrieload", point1=barrierpoint_3, point2=barrierpoint_4
    )
    Barrier3 = og.create_load(
        loadtype="line", name="Barrieload", point1=barrierpoint_5, point2=barrierpoint_6
    )
    Patch1 = og.create_load(
        loadtype="patch",
        point1=barrierpoint_1,
        point2=barrierpoint_3,
        point3=barrierpoint_2,
        point4=barrierpoint_4,
    )

    barrierpoint_1 = og.create_load_vertex(x=6, z=2, p=0)
    barrierpoint_2 = og.create_load_vertex(x=11, z=8, p=0)
    barrierpoint_3 = og.create_load_vertex(x=11, z=3, p=0)
    barrierpoint_4 = og.create_load_vertex(x=6, z=6, p=0)

    Patch2 = og.create_load(
        loadtype="patch",
        point1=barrierpoint_1,
        point2=barrierpoint_3,
        point3=barrierpoint_2,
        point4=barrierpoint_4,
    )
    # create basic load case
    barrier_load_case = og.create_load_case(name="Barrier")
    # barrier_load_case.add_load_groups(Barrier)  # ch
    barrier_load_case.add_load(Patch1)  # ch
    # 2nd
    barrier_load_case2 = og.create_load_case(name="Barrier2")
    # barrier_load_case2.add_load_groups(Barrier2)
    barrier_load_case2.add_load(Patch2)
    # 3rd
    # barrier_load_case3 = og.create_load_case(name="Barrier3")
    # barrier_load_case3.add_load_groups(Barrier3)

    # adding load cases to model
    example_bridge.add_load_case(barrier_load_case)
    example_bridge.add_load_case(barrier_load_case2)
    # example_bridge.add_load_case(barrier_load_case3)

    single_path = og.create_moving_path(
        start_point=og.Point(2, 0, 2), end_point=og.Point(4, 0, 3), increments=3
    )  # create path object
    move_point = og.create_moving_load(name="single_moving_point")
    move_point.set_path(single_path)
    move_point.add_load(load_obj=front_wheel)
    example_bridge.add_load_case(move_point)

    example_bridge.analyze()
    results = example_bridge.get_results(local_forces=False)

    # og.opsv.section_force_diagram_3d("Mz",{})

    f = og.plot_force(
        ospgrillage_obj=example_bridge,
        result_obj=results,
        component="Mz",
        member="interior_main_beam",
        loadcase="single_moving_point at global position [4.00,0.00,3.00]",
    )

    f.show()


def test_shell_plot_force(shell_link_bridge):
    # test functionality of plot_force on a shell_beam model type
    shell_link_model = shell_link_bridge
    # create and add load case comprise of single point load
    P = 20e3
    point_load_location = og.create_load_vertex(
        x=4.5, y=0, z=6.5, p=P
    )  # about midspan of span 1
    point_load = og.create_load(
        loadtype="point", name="single point", point1=point_load_location
    )
    point_lc = og.create_load_case(name="pointload")
    point_lc.add_load(point_load)
    shell_link_model.add_load_case(point_lc)
    shell_link_model.analyze()
    # extract results
    result = shell_link_model.get_results()
    print(result)
    f = og.plot_force(
        ospgrillage_obj=shell_link_model,
        result_obj=result,
        component="Mx",
        member="exterior_main_beam_1",
    )

    f.show()


def test_displacement_getter(bridge_model_42_negative):
    # test functionality of plot_force and its output
    og.ops.wipeAnalysis()
    example_bridge = bridge_model_42_negative
    # og.opsv.plot_model()
    # og.plt.show()

    # add moving load case
    front_wheel = og.PointLoad(
        name="front wheel", point1=og.LoadPoint(7.5, 0, 4.5, 160e3)
    )

    point_load_case = og.create_load_case(name="Point")
    point_load_case.add_load(load_obj=front_wheel)
    example_bridge.add_load_case(point_load_case)

    example_bridge.analyze()
    results = example_bridge.get_results(local_forces=False)

    processor = og.PostProcessor(grillage=example_bridge, result=results)

    arbitrary_disp = processor.get_arbitrary_displacements(point=[5, 0, 3])  # bigger

    assert all(np.isclose(arbitrary_disp * 1e6, [3.1289e-05 * 1e6]))



# -*- coding: utf-8 -*-
"""
This module holds all static methods used in other modules. Static methods are general functions to perform
calculations. Most methods herein are not part of API of ospgrillage - these functions are only called and accessed in
ospgrillage for the purpose of its meshing and calculations. However, users may find some static methods to be useful
for their workflow.
"""

import numpy as np
from scipy.spatial import distance
from scipy.optimize import fsolve, root
import warnings


def diff(li1, li2):
    """
    Static method to determine the difference between two list. used in OspGrillage.set_member() function to check member
    assignment
    """
    return list(list(set(li1) - set(li2)) + list(set(li2) - set(li1)))


def find_circle(x1, y1, x2, y2, x3, y3):
    x12 = x1 - x2
    x13 = x1 - x3

    y12 = y1 - y2
    y13 = y1 - y3

    y31 = y3 - y1
    y21 = y2 - y1

    x31 = x3 - x1
    x21 = x2 - x1

    # x1^2 - x3^2
    sx13 = pow(x1, 2) - pow(x3, 2)

    # y1^2 - y3^2
    sy13 = pow(y1, 2) - pow(y3, 2)

    sx21 = pow(x2, 2) - pow(x1, 2)
    sy21 = pow(y2, 2) - pow(y1, 2)

    f = ((sx13) * (x12) + (sy13) * (x12) + (sx21) * (x13) + (sy21) * (x13)) // (
        2 * ((y31) * (x12) - (y21) * (x13))
    )

    g = ((sx13) * (y12) + (sy13) * (y12) + (sx21) * (y13) + (sy21) * (y13)) // (
        2 * ((x31) * (y12) - (x21) * (y13))
    )

    c = -pow(x1, 2) - pow(y1, 2) - 2 * g * x1 - 2 * f * y1

    # eqn of circle be x^2 + y^2 + 2*g*x + 2*f*y + c = 0
    # where centre is (h = -g, k = -f) and
    # radius r as r^2 = h^2 + k^2 - c
    h = -g
    k = -f
    sqr_of_r = h * h + k * k - c

    # r is the radius
    r = round(np.sqrt(sqr_of_r), 5)

    print("Centre = (", h, ", ", k, ")")
    print("Radius = ", r)

    return [[h, k], r]


# -----------------------------------------------------------------------------------------------------------------------
# Sweep path equations - Add future functions here
# which the given three points lie


def line_func(m=None, c=None, x=None, h=None, v=None, R=None):
    """
    Line function. Returns y position given x.
    """
    curve = False
    if not all([h is None, v is None, R is None]):
        curve = True

    if not curve:
        # straight line
        if type(x) is list:
            y = m * x[0] + c
        else:
            y = m * x + c
    elif curve:
        y = np.sqrt((R) ** 2 - (x - h) ** 2) + v
    else:
        raise Exception(
            "line function missing arguments for valid function selection: check arguments"
        )

    return y


def inv_line_func(m, c, y):
    x = (y - c) / m if all([m != 0, m is not None, c is not None]) else 0
    return x


def arc_func(h, v, R, x, r=0):
    # function to get y coordinate on an arc given the variables of the circle equation
    # option to add radius to circle , lowercase r
    y = np.sqrt((R + r) ** 2 - (x - h) ** 2) + v
    return y


def create_arc_points(point1, radius, length, num_inc):
    # function to create points along arc `length` of a sector with `angle`
    start_angle = np.pi / 2  # 90 degrees
    angle = length / radius  # calculate angle of sector
    end_angle = (
        start_angle - angle
    )  # difference is the end point's angle (about center_point)
    center_point = [point1.x, -radius]  # x z , model plane = 0 default
    # find point2
    angle_inc = np.linspace(start_angle, end_angle, num_inc)
    # point2 = [center_point[0] + radius*np.cos(end_angle), center_point[1] + radius * np.sin(end_angle)]

    x_curve = [center_point[0] + radius * np.cos(ang) for ang in angle_inc]
    z_curve = [center_point[1] + radius * np.sin(ang) for ang in angle_inc]

    return x_curve, z_curve


# -----------------------------------------------------------------------------------------------------------------------
# -----------------------------------------------------------------------------------------------------------------------
def select_segment_function(curve_flag, d, x, r=0, m=0, c=0):
    if curve_flag:
        y = arc_func(h=d[0][0], v=d[0][1], R=d[1], x=x, r=r)
    else:
        y = line_func(m, c, x)
    return y


def find_dict_key(my_dict, key):
    return eval(list(my_dict.keys())[list(my_dict.values()).index(key)])


def x_intcp_two_lines(m1, m2, c1, c2):
    x = (c2 - c1) / (m1 - m2)
    return x


def line(p1, p2):
    A = p1[1] - p2[1]
    B = p2[0] - p1[0]
    C = p1[0] * p2[1] - p2[0] * p1[1]
    return A, B, -C


def intersection(L1, L2):
    D = L1[0] * L2[1] - L1[1] * L2[0]
    Dx = L1[2] * L2[1] - L1[1] * L2[2]
    Dy = L1[0] * L2[2] - L1[2] * L2[0]
    if D != 0:
        x = Dx / D
        y = Dy / D
        return x, y
    else:
        return False


def get_y_intcp(m, x, y):
    c = y - x * m if m is not None else None
    return c


def get_line_func(skew_angle, node_point):
    m = 1 / np.tan(-skew_angle / 180 * np.pi)
    if len(node_point) < 3:
        c = get_y_intcp(m=m, x=node_point[0], y=node_point[1])
    else:
        c = get_y_intcp(m=m, x=node_point[0], y=node_point[2])
    return m, c


def find_min_x_dist(const_point, ref_point):
    # constant point is node point on construction line
    # ref_point is roving point on the arbitrary line
    d = distance.cdist(const_point, ref_point)
    return d


def get_slope(pt1, pt2):
    # pt must be [x y z]
    # function claculates the slope for two points inthe 2-D plane, y= 0
    if (pt1[0] - pt2[0]) == 0:
        m = None
    else:
        m = (pt1[2] - pt2[2]) / (pt1[0] - pt2[0])

    # find phi angle
    if m is None:
        phi = np.pi / 2
    else:
        phi = np.arctan(m)
    return m, phi


def solve_zeta_eta(xp, zp, x1, z1, x2, z2, x3, z3, x4, z4):
    # create function to solve for eta and zeta - dynamically for varying parameters x1-x4, z1-z4, zp,xp

    # mapping of natural coordinate eta{-1:1}, zeta{-1:1} to global coordinate (x,z)
    def obj_func(x, xp, zp, x1, x2, x3, x4, z1, z2, z3, z4):
        eta = 4 * zp - (
            (1 - x[0]) * (1 - x[1]) * z1
            + (1 + x[0]) * (1 - x[1]) * z2
            + (1 + x[0]) * (1 + x[1]) * z3
            + (1 - x[0]) * (1 + x[1]) * z4
        )
        zeta = 4 * xp - (
            (1 - x[0]) * (1 - x[1]) * x1
            + (1 + x[0]) * (1 - x[1]) * x2
            + (1 + x[0]) * (1 + x[1]) * x3
            + (1 - x[0]) * (1 + x[1]) * x4
        )
        return eta, zeta

    # if initial xp zp is very close (order 1e-8) to the centre (0,0,0) of natural coordinate,
    # avoid fsolve RunTimeWarning by
    if all([np.isclose(xp, (x1 + x2) / 2), np.isclose(zp, (z1 + z3) / 2)]):
        root_func = [0, 0]

    else:
        root_func = fsolve(
            obj_func,
            np.array([-0.4444444444444, 0.99999999]),
            args=(xp, zp, x1, x2, x3, x4, z1, z2, z3, z4),
        )  # here the initial values are chosen to avoid RunTimeWarning where initial guess is very close
        # to the final solution of eta and zeta

    eta = root_func[0]
    zeta = root_func[1]
    return eta, zeta


def get_distance(a, b):
    return np.sqrt((a.x - b.x) ** 2 + (a.z - b.z) ** 2)


def is_between(a, c, b):
    return get_distance(a, c) + get_distance(c, b) == get_distance(a, b)


def calculate_area_given_vertices(p_list):
    # input list of namedtuple LoadPoint or Point
    # note: p_list must have been sorted in counter clockwise via sort_vertices function
    # (this is called prior to this function)
    x_list = []
    y_list = []
    z_list = []
    A_mag = 0
    for point in p_list:
        x_list.append(point.x)
        y_list.append(point.y)
        z_list.append(point.z)
    # add first point to list
    x_list.append(p_list[0].x)
    y_list.append(p_list[0].y)
    z_list.append(p_list[0].z)
    for count in range(len(x_list) - 1):
        A_mag += x_list[count] * z_list[count + 1] - x_list[count + 1] * z_list[count]
    A = np.abs(A_mag / 2)
    return A


def check_point_in_grid(inside_point, point_list):
    # ref: solution 3 https://www.eecs.umich.edu/courses/eecs380/HANDOUTS/PROJ2/InsidePoly.html
    # check counter clockwise and to the left (greater than 0)
    # put points in list
    pt0 = point_list  # list of point tuples
    # rotate point
    pt1 = pt0[1:] + [pt0[0]]
    inside = True
    if inside_point.z is None:
        inside = False
        return inside
    # check if point is clockwise via sign of signed_area
    signed_area = check_points_direction(
        point_list
    )  # sign area < 0 means points are clockwise, and vice versa
    for count, point0 in enumerate(pt0):
        side = (inside_point.z - point0.z) * (pt1[count].x - point0.x) - (
            inside_point.x - point0.x
        ) * (pt1[count].z - point0.z)
        # check if point is outside
        if any(
            [
                side
                < 0
                <= signed_area,  # side > 0 means left, side < 0 means right, side = 0 means on the line path
                side > 0 > signed_area,
            ]
        ):  # nodes are counterclockwise and point to right (outside)
            inside = False  # nodes are clockwise and point to the left (outside) these condition
    return inside  # return line = outside


def check_points_direction(point_list):
    # list of point tuples
    # ref http://mathworld.wolfram.com/PolygonArea.html
    signed_area = 0
    last_point = None
    first_point = None
    x2 = None
    z2 = None
    for counter, point in enumerate(point_list):
        x1 = point.x
        z1 = point.z
        if first_point is None:
            # save point as first point
            first_point = point

        if counter == len(point_list) - 1:
            x2 = first_point.x
            z2 = first_point.z
        else:
            x2 = point_list[counter + 1].x
            z2 = point_list[counter + 1].z
        signed_area += x1 * z2 - x2 * z1

        last_point = point  # set current point as last point
    return signed_area


# ----------------------------------------------------------------------------------------------------------
# function to check intersection of line segments
def onSegment(p, q, r):
    # point nameTuple p, q and r
    if (
        (q.x <= max(p.x, r.x))
        and (q.x >= min(p.x, r.x))
        and (q.y <= max(p.z, r.z))
        and (q.z >= min(p.z, r.z))
    ):
        return True
    return False


def orientation(p, q, r):
    # to find the orientation of an ordered triplet (p,q,r)
    # function returns the following values:
    # 0 : Colinear points
    # 1 : Clockwise points
    # 2 : Counterclockwise

    # See https://www.geeksforgeeks.org/orientation-3-ordered-points/amp/
    # for details of below formula.
    # modified - p, q, r must be point namedtuple
    val = (float(q.z - p.z) * (r.x - q.x)) - (float(q.x - p.x) * (r.z - q.z))
    if val > 0:
        # Clockwise orientation
        return 1
    elif val < 0:
        # Counterclockwise orientation
        return 2
    else:
        # Colinear orientation
        return 0


# Function returns true if the line segment 'p1q1' and 'p2q2' intersect.
def check_intersect(p1, q1, p2, q2):
    # ref https://www.geeksforgeeks.org/check-if-two-given-line-segments-intersect/
    general_intersect = False
    colinear = False
    # Find the 4 orientations required for
    # the general and special cases
    o1 = orientation(p1, q1, p2)
    o2 = orientation(p1, q1, q2)
    o3 = orientation(p2, q2, p1)
    o4 = orientation(p2, q2, q1)

    # General case
    if (o1 != o2) and (o3 != o4):
        general_intersect = True
        colinear = False
        return general_intersect, colinear

    # Special Cases
    # p1 , q1 and p2 are colinear and p2 lies on segment p1q1
    if (o1 == 0) and onSegment(p1, p2, q1):
        general_intersect = True
        colinear = True
        return general_intersect, colinear

    # p1 , q1 and q2 are colinear and q2 lies on segment p1q1
    if (o2 == 0) and onSegment(p1, q2, q1):
        general_intersect = True
        colinear = True
        return general_intersect, colinear

    # p2 , q2 and p1 are colinear and p1 lies on segment p2q2
    if (o3 == 0) and onSegment(p2, p1, q2):
        general_intersect = True
        colinear = True
        return general_intersect, colinear

    # p2 , q2 and q1 are colinear and q1 lies on segment p2q2
    if (o4 == 0) and onSegment(p2, q1, q2):
        general_intersect = True
        colinear = True
        return general_intersect, colinear

    # If none of the cases
    return general_intersect, colinear


# --------------------------------------------------------------------------------------------
def find_plane_centroid(point_list):
    """
    Function to find centroid given a set of vertices (for patch load). function calculates 2D plane x z - y plane is
    the model plane.
    :param point_list: list of Point() namedtuples
    :type point_list: list
    """
    x = 0
    z = 0
    for point in point_list:
        x += point.x
        z += point.z
    x_c = x / len(point_list)
    z_c = z / len(point_list)
    return [x_c, z_c]


def sort_vertices(point_list, node_tag_list=None):
    # note the node_tag_list must correspond to that of point_list, this is ensure in the higher level
    # functions which calls sort_vertices
    if node_tag_list is None:
        node_tag_list = []
    center_x_z = find_plane_centroid(point_list)
    angle_list = []
    for point in point_list:
        # calculate angle
        x = np.array(point.x - center_x_z[0])
        z = np.array(point.z - center_x_z[1])
        angle_list.append(np.arctan2(z, x))
        # angle_list.append(np.arctan((point.z - center_x_z[1]) / (point.x - center_x_z[0])))
    # angle_list[angle_list<0] = angle_list[angle_list<0] + 2*np.pi
    [i + 2 * np.pi for i in angle_list if i < 0]
    # sort for counter clockwise
    sorted_points = [x for _, x in sorted(zip(angle_list, point_list))]

    sorted_node_tag = [x for _, x in sorted(zip(angle_list, node_tag_list))]
    return sorted_points, sorted_node_tag


def get_patch_centroid(point_list):
    # function to find centroid on a plane (x-z) accounting for separate value of area load on each node point
    m_total = []
    mx = 0
    my = 0
    mz = 0
    for point in point_list:
        m_total.append(point.p)
        mx += point.x * point.p
        mz += point.z * point.p
        my += point.y * point.p
    sum_m = sum(m_total) if sum(m_total) > 0 else 1
    xc = mx / sum_m
    zc = mz / sum_m
    yc = my / sum_m
    return xc, yc, zc


# abstracted function for assigning patch loading
def check_dict_same_keys(d_1, d_2):
    merged = {**d_1, **d_2}
    # function to check if two dicts have same key
    same_key = [k in d_2 for k in d_1.keys()]
    if any(same_key):
        common_key = [k for (k, v) in zip(d_1, same_key) if v]
        # get val from d_1 and d_2
        for grid in common_key:  # key is grid number
            first_list = d_1[grid]
            second_list = d_2[grid]
            updated_list = dict()
            for key, val_1 in first_list.items():
                # if key == 'ends':
                #     val_end = val_1 if val_1 else second_list.get(key)
                #     updated_list.setdefault(key, val_end)
                # else:
                combine_val = []
                val_2 = second_list.get(key)
                for val in val_1 + val_2:
                    if val not in combine_val:
                        combine_val.append(val)

                updated_list.setdefault(key, combine_val)
            merged.update({grid: updated_list})

    return merged


def sort_list_into_four_groups(group_list: list, option: str = None):
    """
    Function to sort a list of group number into four groups , returns a dict with keys [0,1,2,3] placing the ordered
    group_list into each key. Purpose is to distinguish main grillage elements based on the group positioning
    e.g. if group_list = [0,1,2,3,4,5,6], returns {0:[0,6], 1:[1],2:[2,3,4],3:[5]}
    :param option: type of group sorting - default None for beam, "shell" for shell model
    :type option: str
    :param group_list:
    :return: dict with keys [0,1,2,3] and values be the sorted group list as shown in description
    """
    # sort asccending
    output_dict = dict()
    if option == "shell":
        # add two proxy element at start and end of group list
        group_list = [group_list[0] - 1] + group_list + [group_list[-1] + 1]

    # edges
    output_dict[0] = [group_list[0], group_list[-1]]
    # instantiate
    exterior_beam_group_1 = [group_list[1]]
    exterior_beam_group_2 = [group_list[-2]]
    interior_beam_group = group_list[2 : len(group_list) - 2]

    # check for special cases
    if len(group_list) > 2:
        if len(group_list) == 3:  # 2 edge and 1 interior
            interior_beam_group = [group_list[1]]
            exterior_beam_group_1 = []
            exterior_beam_group_2 = []
        elif len(group_list) == 4:  # 2 edge, ext_1 and ext_2
            interior_beam_group = []

    output_dict[1] = exterior_beam_group_1
    output_dict[2] = interior_beam_group
    output_dict[3] = exterior_beam_group_2
    return output_dict


def rotate_point_about_point(center_x, center_y, angle, point: list):
    """
    Rotates point about a second point at [center_x, center_y] by angle in the coordinate system

    param point: list containing the x and y coordinate
    type point: list
    param angle: Angle in radians
    type angle: int or float
    returns: list of coordinate for the rotated point
    """

    s = np.sin(angle)  # rad
    c = np.cos(angle)  # rad
    if isinstance(point, list):
        x_or = point[0] - center_x
        y_or = point[1] - center_y

        x_rot = x_or * c - y_or * s
        y_rot = x_or * s + y_or * c

        # translate back to origin and store to var
        rotated_point = [x_rot + center_x, y_rot + center_y]
    else:
        raise Exception("point= requires a valid list or tuple")

    return rotated_point



from ospgrillage.utils import *
from ospgrillage.mesh import *
from ospgrillage.load import *
from ospgrillage.material import *
from ospgrillage.members import *
from ospgrillage.osp_grillage import *
from ospgrillage.postprocessing import *

__version__ = "0.4.0"
